#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include "brush.h"
#include "buffer.h"

static buffer_t *sbuf;

static void begin_stroke()
{
  sbuf= buffer_new(50);
}

static void end_stroke(stroke_t *S)
{
  size_t rlen;
  buffer_free(sbuf, (void**) &S->data, &rlen);
  S->len= rlen/sizeof(S->data[0]);
}

static void stroke_point(int x,int y,unsigned int value)
{
  relp_t K;
  K.value= value;
  K.dx= x;
  K.dy= y;
  buffer_write(sbuf, &K, sizeof(K));
}

static brush_t* brush_create(unsigned char *data, int W, int H)
{
  brush_t *B;
  int cx, cy, x, y, fx, fy, tx, ty, ddx, ddy;
  unsigned int fv, tv;
  int s;

  static int DX[8]= {  0,  1,  1,  1,  0, -1, -1, -1 },
             DY[8]= { -1, -1,  0,  1,  1,  1,  0, -1 };

  cx= W/2; cy= H/2;

  B= calloc(1,sizeof(*B));

  begin_stroke();
  for(y=0;y<H;y++) 
    for(x=0;x<W;x++)
      if (data[x+y*W]) stroke_point(x-cx,y-cy, data[x+y*W]);
  end_stroke(B->strokes+0);

  for(s=0;s<8;s++)
  {
    ddx= DX[s]; ddy= DY[s];
    begin_stroke();
    for(ty=0;ty<H;ty++) for(tx=0;tx<W;tx++)
    {
      fx= tx - ddx; 
      fy= ty - ddy;
      if (fx<0 || fy<0 || fx>=W || fy>=H) fv= 0;
                  else  fv= data[fx+fy*W];
      tv= data[tx+ty*W];
      if (tv>=fv) continue;
      stroke_point(fx-cx, fy-cy, fv);
    }
    end_stroke(B->strokes+s+1);
  }
  return B;
}

brush_t* brush_filled_circle(int R,int S)
{
  int x,y,sy,sx,total, cx,cy,tx,ty, ddx, ddy, fx, fy;
  int s;
  unsigned char *data, pix;
  brush_t *B;
  int D;

  if (R%2) D= R+15;
      else D= R+14; 
  data= malloc(D*D);

  cx= cy= D*S/2 + S/2;

  for(y=0;y<D;y++) for(x=0;x<D;x++) // for each pixel
  {
    total= 0;
    for(sy=0;sy<S;sy++) for(sx=0;sx<S;sx++) // for each subpixel
    {
       ty= (y*S+sy) - cx; tx= (x*S+sx)-cy;
       if (4*(tx*tx+ty*ty)<=R*R*S*S) total++;
    }
    pix= total*255 / (S*S);
    if (pix>255) pix= 255;
    data[y*D+x]= pix; 
  }

  B= brush_create(data, D, D);
  free(data);
  return B;
}

