#include <stdlib.h>
#include <stdint.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include "gfx.h"
#include "draw.h"
#include "brush.h"

static int wi_pix;
static unsigned int *img, iw,ih;

static int uminx, uminy, umaxx, umaxy;

static int imin2(int a,int b) { return a<b?a:b; }
static int imax2(int a,int b) { return a>b?a:b; }

static brush_t *brush;
static int bwidth;
static unsigned int fg;

static void update(int x,int y,int w,int h)
{
  gfx_pixgrid_put(wi_pix, x,y,img, iw,ih,x,y,w,h);
}

static void begin_paint()
{
  uminx= iw; uminy= ih; umaxx=0; umaxy= 0;
}

static void end_paint()
{
  int x,y, w,h;
  uminx= imax2(0,uminx); uminy= imax2(0,uminy);
  umaxx= imin2(iw-1, umaxx); umaxy= imin2(ih-1,umaxy);
  x= uminx; y= uminy;
  w= umaxx-uminx+1; h= umaxy-uminy+1;
  if (w<=0 || h<=0) return ;
  update(x,y,w,h);
}

void draw_init(int widget)
{
  int dum;
  wi_pix= widget;
  gfx_pixgrid_attribs(wi_pix, &dum, &iw, &ih);

  img= malloc(sizeof(*img)*iw*ih);
  memset(img, 0, iw*ih*sizeof(*img));

  update(0,0,iw,ih);                  
}

static int coor2pix(int *x,int *y)
{
  *x= (*x+100)/200.0*iw;
  *y= (*y+100)/200.0*ih;
  if (*x<0 || *y<0 || *x>=iw || *y>=iw) return 1;
  return 0;
}

void draw_width(int w)
{
  brush= brush_filled_circle(w,4); 
  bwidth= w;
}

void draw_color(unsigned int color)
{
  fg= color;
}

static unsigned int blend(unsigned int A, unsigned int F, unsigned int B)
{
  unsigned int fR, fG, fB, bR, bG, bB, pR, pG, pB;
  fR= (F>>16)&0xff; fG= (F>>8)&0xff; fB= F&0xff;
  bR= (B>>16)&0xff; bG= (B>>8)&0xff; bB= B&0xff;
  pR= (fR*A+bR*(255-A)) / 255;
  pG= (fG*A+bG*(255-A)) / 255;
  pB= (fB*A+bB*(255-A)) / 255;
  return (pR<<16) | (pG<<8) | (pB);
}

static void stroke(int x0,int y0,stroke_t *S)
{
  int  x,y;
  int i;
  for(i=0;i<S->len;i++)
  {
    x= x0+ S->data[i].dx; if (x<0 || x>=iw) continue;
    y= y0+ S->data[i].dy; if (y<0 || y>=ih) continue;
    img[x+y*iw]= blend(S->data[i].value, fg, img[x+y*iw]);
    uminx= imin2(x,uminx); umaxx= imax2(x,umaxx);
    uminy= imin2(y,uminy); umaxy= imax2(y,umaxy);
  }
}

static void draw_line_singlepix(int x1,int y1,int x2, int y2)
{
  int dx,dy, x,y, stepx, stepy, e,
  nstepx, nstepy, dstepy, dstepx, Ldia, Lnor, Lthr;

  coor2pix(&x1,&y1); coor2pix(&x2,&y2);
  dx= x2-x1; stepx= dx<0? -1: 1; dx= dx<0? -dx:dx;
  dy= y2-y1; stepy= dy<0? -1: 1; dy= dy<0? -dy:dy;


  begin_paint();

  int d, p;
  d= imax2(dx,dy);
  p= imin2(dx,dy);
  Ldia= 2*d;
  Lnor= 2*p;
  Lthr= d;
  if (dx>=dy) {  nstepx= stepx; nstepy= 0; dstepy= stepy; dstepx= 0; }
         else {  nstepy= stepy; nstepx= 0; dstepx= stepx; dstepy= 0; }

  for(e=0,x=x1,y=y1;d;d--,x+=nstepx,y+=nstepy,e+=Lnor)
  {
    if (x>=0 && x<iw && y>=0 && y<ih) img[x+y*iw]= fg;
    if (e+Lnor>=Lthr) { x+= dstepx; y+= dstepy; e-= Ldia; }
  }
  uminx= imin2(x1,x2); umaxx= imax2(x1,x2);
  uminy= imin2(y1,y2); umaxy= imax2(y1,y2);
  end_paint();
}

void draw_line(int x1,int y1,int x2, int y2)
{
  int dx,dy, x,y, t, stepx, stepy, e;

  if (bwidth<2) { draw_line_singlepix(x1,y1,x2,y2); return ; }
  coor2pix(&x1,&y1); coor2pix(&x2,&y2);
  dx= x2-x1; stepx= dx<0? -1: 1; dx= dx<0? -dx:dx;
  dy= y2-y1; stepy= dy<0? -1: 1; dy= dy<0? -dy:dy;

  begin_paint();

  stroke_t *dia, *nor;
  
  if (dx>=dy)
  {
    if (x1>x2) { t=x1; x1=x2; x2=t; t=y1; y1=y2; y2=t;
                stepy=-stepy; stepx=-stepx;}
    nor= brush->strokes+3;
    if (y1==y2)  dia= brush->strokes+3; 
    else if (y1<y2) dia= brush->strokes+4; 
    else dia= brush->strokes+2;
    e= 0;
    y= y1;
             //stroke(x1,y,brush->strokes+0); goto done;
    //stroke(x1,y1,brush->strokes+2);  goto done;
    stroke(x1,y1,brush->strokes+0); 
    e= 2*dy;
    for(x=x1;x<=x2;x++)
    {
       if (e+2*dy>dx) { stroke(x,y, dia); y+= stepy; e-= 2*dx; }
       stroke(x,y,nor);
       e+= 2*dy;
    } 
  }
  else
  {
    if (y1>y2) { t=x1; x1=x2; x2=t; t=y1; y1=y2; y2=t;
                stepy=-stepy; stepx=-stepx;}
    nor= brush->strokes+5;
    if (x1==x2)  dia= brush->strokes+5; 
    else if (x1<x2) dia= brush->strokes+4; 
    else dia= brush->strokes+6;
    e= 0;
    x= x1;
    stroke(x1,y1,brush->strokes+0); 
    e= 2*dy;
    for(y=y1;y<=y2;y++)
    {
       if (e+2*dx>dy) { stroke(x,y, dia); x+= stepx; e-= 2*dy; }
       stroke(x,y,nor);
       e+= 2*dx;
    } 
  } 

  end_paint();
}


void draw_clear()
{
  int i,T;
  T= iw*ih;
  for(i=0;i<T;i++) img[i]= 0;
  update(0,0,iw,ih);
}

static void circ_pix(int x,int y)
{
  if (bwidth>2) { stroke(x,y,brush->strokes+0); return ; }
  if (x<0 || y<0 || x>=iw || y>=iw) return ;
  img[x+y*iw]= fg;
  uminx= imin2(x,uminx);
  uminy= imin2(y,uminy);
  umaxx= imax2(x,umaxx);
  umaxy= imax2(y,umaxy);
}

void draw_circle(int xc, int yc, int r)
{
  int p,x,y;
  begin_paint();
  coor2pix(&xc,&yc);
  r= r * iw / 200;
  for(x=0,y=r,p=3-2*r;x<=y;x++)
  {
    if (p<0)  p= p+4*x+6; 
       else { y--; p= p+4*(x-y)+10; }
    circ_pix(xc+x,yc-y);
    circ_pix(xc-x,yc-y);
    circ_pix(xc+x,yc+y);
    circ_pix(xc-x,yc+y);
    circ_pix(xc+y,yc-x);
    circ_pix(xc-y,yc-x);
    circ_pix(xc+y,yc+x);
    circ_pix(xc-y,yc+x);
  }
  end_paint();
}
