#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "editor.h"
// This will be a really simple editor. It will load everything into one
// big chunk, and then make rows out of it. 
// each row will end with a new line, which isn't stored in the 
// editor buffer.
// 

static int pageH, rowW;

typedef struct row
{
  int line;
  int drow;
  char *data; 
  struct row *prev, *next;
} row_t;

static row_t *rows_head;

typedef struct { row_t *row; int start, len; } upd_t;

static int nupdates, supdates;
static upd_t *updates;

static row_t *Crow;
static int Ccol;


static void update(row_t *R, int start, int len)
{
  upd_t *U;
  if (!R)
  {
    // printf("hohoho! null row update??\n");
    *(int*)0=0;
  }
  if (nupdates==supdates)
  {
    supdates += supdates/2 + 4;
    updates= realloc(updates, sizeof(updates[0])*supdates);
  }
  U= updates+nupdates;
  U->row= R;
  U->start= start;
  U->len= len;
  nupdates++;
}

static void run_upd(row_t *R, int start, int len)
{
  editor_print(R->drow, start, 0, start+R->data, len);
  if (R==Crow && start<=Ccol && start+len>Ccol)
    editor_print(Crow->drow, Ccol, 1, Crow->data+Ccol,1);
}

static void make_cursor_visible()
{
  int Y;
  row_t *K;
  for(K=Crow->prev;K;K=K->prev) if (K->drow!=-1) break;
  if (K)
  {
    Y= pageH-1;
    for(K=Crow;K;K=K->prev)  
     { if (Y==-1) break;  K->drow= Y; run_upd(K,0,rowW); Y--; }
    for(;K;K=K->prev) { if (K->drow==-1) break; K->drow= -1; }
    return ;
  }
  Y= 0;
  for(K=Crow;K;K=K->next) 
   { if (Y==pageH) break; K->drow= Y; run_upd(K,0,rowW); Y++; }
  for(;K;K=K->next) { if (K->drow==-1) break; K->drow= -1; }
}

static void do_all_updates()
{
  upd_t *U;
  int i;

  if (Crow->drow==-1) { make_cursor_visible(); nupdates= 0; return ; }

  for(i=0;i<nupdates;i++)
  {
    U= updates+i;
    if (U->row->drow==-1) continue;
//  printf("update[row %d, col=%d, len=%d]\n", U->row->drow, U->start, U->len);
    run_upd(U->row, U->start, U->len);
  }
  nupdates= 0;
}



static void rowins(row_t *R, int pos, int chr)
{
  int i;
  for(i=rowW-1;i>pos;i--) R->data[i]= R->data[i-1];
  R->data[pos]= chr;
  update(R, pos, rowW-pos);
}

static row_t *row_alloc(int line, int drow)
{
  row_t *N;
  N= calloc(1,sizeof(*N));
  N->line= line;
  N->drow= drow;
  N->data=calloc(1,rowW);
  return N;
}

static void shift_rows_down(row_t *A)
{
  row_t *K;
  for(K=A;K;K=K->next) 
  { 
    if (K->drow==-1) break; 
    K->drow++; 
    if (K->drow>=pageH) K->drow= -1; 
                   else update(K,0,rowW);
  }
}

static void shift_rows_up(row_t *A)
{
  row_t *K;
  for(K=A;K;K=K->next)
  {
    if (K->drow==-1) { K->drow= pageH-1; update(K,0,rowW); break; }
    K->drow--;
    if (K->drow!=-1) update(K,0,rowW);
  }
}

static void insert_row_after(row_t *H, row_t *N)
{
  N->next= H->next;
  N->prev= H;

  H->next= N;
  if (N->next) N->next->prev= N;
}


static row_t *alloc_cont_row(row_t *R)
{
  row_t *N;
  N= row_alloc(R->line, R->drow+1);
  shift_rows_down(R->next);
  insert_row_after(R, N);
  return N;
}

static row_t *get_cont_row(row_t *R)
{
  row_t *N;
  N= R->next;
  if (!N) return NULL;
  if (N->line!=R->line) return NULL;
  return N;
}

static row_t *get_or_alloc_cont_row(row_t *R)
{
  row_t *N;
  N= get_cont_row(R);
  if (N) return N;
  return alloc_cont_row(R);
}

static void editor_insert_char(int chr)
{
  int i,oc;
  row_t *R; int pos;
  R= Crow; pos= Ccol;

  if (R->data[pos]==0)  // past the end of present characters.
  {
    R->data[pos]= chr;
    for(i=0;i<pos;i++) if (R->data[i]==0) R->data[i]= ' ';
    update(R, pos, 1); 
// printf("FirsT!\n");
    goto done;
  }
again:
  if (R->data[rowW-1]==0) // we have space for the char, in this row.
  {
    rowins(R, pos, chr);
    goto done;
  }
  // wrap around. we make a new row, or overflow this row to the next one.

  oc= R->data[rowW-1];
  for(i=rowW-1;i>pos;i--) R->data[i]= R->data[i-1];
  R->data[pos]= chr;
  update(R, pos, rowW-pos);

  R= get_or_alloc_cont_row(R);
  pos= 0;
  chr= oc;
  goto again;
done:
  if (Ccol!=rowW-1)
  {
    Ccol= Ccol+1;
  }
  else
  {
    Crow= get_or_alloc_cont_row(Crow);
    Ccol= 0;
  }
  update(Crow, Ccol,1);
  do_all_updates();
}

void editor_init(int nrows,int ncols)
{
  row_t *R,*P;
  int L,i;

  pageH= nrows;
  rowW= ncols; 
  P= NULL;
  L= 0;
  rows_head= NULL;
  nupdates= 0;
  supdates= 2*pageH;
  updates= malloc(sizeof(updates[0])*supdates);

  for(i=0;i<pageH;i++)
  {
    R= row_alloc(L,L);
    if (P) P->next= R;
      else rows_head= R;
    R->prev= P;
    R->drow= L;
    P= R;
    L++;
    update(R, 0, rowW);
  }

  Crow= rows_head;
  Ccol= 0;
  do_all_updates();
}

static void unlink_row(row_t *R)
{
  if (R->prev) R->prev->next= R->next;
  if (R->next) R->next->prev= R->prev;
  R->drow= -1;
}

static int count_chars(row_t *R)
{
  int i,c;
  c= 0;
  for(i=0;i<rowW;i++) if (R->data[i]) c++;
  return c;
}

static void simple_backspace(row_t *R, int col)
{
  row_t *N;
  int i;

again:
  for(i=col;i<rowW;i++) R->data[i-1]= R->data[i];
  update(R,0,rowW);
  N= get_cont_row(R);
  if (!N)
  {
     R->data[rowW-1]= 0;
     return ;
  }
  R->data[rowW-1]= N->data[0];
  if (count_chars(N)>1)
  {
    R= N;
    col= 1;
    goto again;
  }
  unlink_row(N);
  shift_rows_up(N->next);
  free(N->data);
  free(N);
}

static void propagate_chars_up(row_t *R)
{
  row_t *N;
  int avail, cr, copied,i;
again:
  N= get_cont_row(R);
  if (!N) return ;
  cr= count_chars(R);
  avail= rowW-cr;
  if (avail==0) { update(R,0,rowW); R=N; goto again; }
  for(i=0;i<avail;i++) R->data[cr+i]= N->data[i]; 
  update(R, 0, rowW);
  copied= count_chars(R) - cr;
  for(i=0;i+copied<rowW;i++) N->data[i]= N->data[i+copied];
  for(i=0;i<copied;i++) N->data[rowW-1-i]= 0;
  if (get_cont_row(N))
  {
    R= N;
    goto again;
  }
  if (count_chars(N)!=0) 
  {
    update(N,0, rowW);
    return ;
  }
  unlink_row(N);
  shift_rows_up(N->next);
  free(N->data); free(N);
}

static void editor_backspace()
{
  row_t *P,*K;
  int cp;
  if (Ccol)
  {
    simple_backspace(Crow, Ccol);
    Ccol--;
    do_all_updates();
    return ;
  }
  P= Crow->prev;
  if (!P) return ;
  if (P->line!=Crow->line) for(K=Crow;K;K=K->next) K->line--;
  cp= count_chars(P);
  if (cp<=1)
  {
    unlink_row(P);
    free(P->data); 
    free(P);
    shift_rows_up(Crow);
    do_all_updates();
    return ;
  }
  P->data[cp-1]= 0; // deleted the last character of the previous row
  // now lets propagate characters from following rows into this one
  // filling it up and moving forward as long as we have more rows
  // for the same line. 
  propagate_chars_up(P);
  Crow= P;
  Ccol= cp-1;
  do_all_updates();
  return ;
}

static void make_space(row_t *R, int nc, char* start)
{
  int i, avail;
  row_t *N;
again:
  for(i=rowW-1;i>=0;i--) if (R->data[i]!=0) break;
  avail= rowW-i-1;

  if (avail>=nc)
  {
    for(i=rowW-1;i>=nc;i--) R->data[i]= R->data[i-nc];
    for(i=0;i<nc;i++) { R->data[i]= start[i]; start[i]= 0; }
    update(R, 0, rowW);
    return ;
  }
  N= get_or_alloc_cont_row(R);
  make_space(N, nc-avail, R->data+rowW-nc);
  goto again;
}

static void editor_break_line()
{
  row_t *N,*K;
  int i,nc;

  N= get_cont_row(Crow);
  if (!N)  goto no_cont_row;

  // we have a continuation row. we combine the rest of this row with the
  // contents in the next row. the next row then starts a new line. 
  // in order to do this, I'll simply make some space in the next row
  // and then copy the rest of this line into there. 
  for(K=N;K;K=K->next) K->line++;

  for(i=Ccol;i<rowW;i++) if (Crow->data[i]==0) break;
  nc= i-Ccol; // number of characters to move.
  make_space(N, nc, Crow->data+Ccol);
  update(Crow, Ccol, rowW-Ccol);

  Crow= N;
  Ccol= 0;
  goto done;

no_cont_row: 
     // the cursor is on the last row of the line. we move the rest of the
     // row into the new line and be done with it.
  N= row_alloc(Crow->line+1, Crow->drow);
  insert_row_after(Crow, N);
  for(K=N->next;K;K=K->next) K->line++;
  shift_rows_down(N);
  if (Crow->data[Ccol]==0) 
  { 
    update(Crow, Ccol,1);
  }
  else
  {
    for(i=Ccol;i<rowW;i++) 
    {
      N->data[i-Ccol]= Crow->data[i];
      Crow->data[i]= 0; 
    }
    update(Crow, Ccol, rowW-Ccol);
  }
  Crow= N;
  Ccol= 0;
  update(N, 0, rowW);

done:
  do_all_updates();
}


static void page_layout(row_t *S)
{
  int i;
  int count;
  row_t *T;
  for(count=0,T=S;T && count<pageH;count++,T=T->next) ;
  if (count!=pageH)
  {
    T= S;
    while(count<pageH)
    { if (!T->prev) break; T=T->prev; count++; }
    S= T;
  }
  for(i=0;S&&i<pageH;i++,S=S->next)
  {
    S->drow= i;
    run_upd(S,0,rowW);
  }
}
static void cursor_move(row_t *R, int col)
{
  if (!R) return ;
  if (col<0) { col= rowW-1; R= R->prev; if (!R) return ; }
  if (col>=rowW) { col= 0; R= R->next; if (!R) return ; }
  update(Crow, Ccol,1);
  Crow= R;
  Ccol= col;
  update(Crow, Ccol,1);
  do_all_updates();
}

static int clear_drows()
{
  int d;
  row_t *K;
  d= Crow->drow;
  if (d==-1) d= 0;
  for(K=Crow->prev;K;K=K->prev) { if (K->drow==-1) break; K->drow= -1; }
  for(K=Crow;K;K=K->next) { if (K->drow==-1) break; K->drow= -1; }
  return d;
}

static void editor_page_up()
{
  int i,K;
  row_t *P,*L,*A;
  if (!Crow->prev) return ;
  K= clear_drows();
  for(A=Crow->prev,i=0;A && i<pageH+K;i++,A=A->prev) L=A;
  if (!A) A= L;
  for(P=A,i=0;i<K&&P;i++,P=P->next) ;
  if (P) Crow= P; else Crow= A;
  page_layout(A);
}

static void editor_page_down()
{
  int i,K;
  row_t *P,*L,*A;
  if (!Crow->next) return ;
  K= clear_drows();
  for(A=Crow->next,i=0;A && i<pageH+K;i++,A=A->next) L=A;
  if (!A) A= L;
  for(P=A,i=0;i<K&&P;i++,P=P->prev) ;
  if (P) Crow= P; else Crow= A;
  page_layout(A);
}

int editor_save(char *fn)
{
  FILE *f;
  row_t *H,*T,*R;
  if (!fn || !Crow) return 1;
  f= fopen(fn,"wb");
  if (!f) return 1;

  for(H=Crow;H->prev;H=H->prev) ;
  for(T=Crow;T->next;T=T->next) ;

  while(T!=H && count_chars(T)==0) T=T->prev;

  for(R=H;R!=T->next;R=R->next)
  {
    int c;
    if (R->prev && R->line!=R->prev->line) fputc('\n',f);
    c= count_chars(R);
    if (c) fwrite(R->data,1,c,f);
  }
  fputc('\n',f);
  fclose(f);
  return 0;
}

int editor_load(char *fn)
{
  FILE *f;
  row_t *R,*H,*T;
  int line, Rl,i;// ,cp,adv,L;
  char buf[200];
  int BL;
  int nrows;

  f= fopen(fn,"rb");
  if (!f) return 1;

  H= T= R= row_alloc(0, -1);
  line= 0;
  Rl= 0;
  nrows= 1;
  
again:
  BL= fread(buf, 1, sizeof(buf), f);
  if (BL<=0) goto end_of_file;
  for(i=0;i<BL;i++)
  {
    if (buf[i]=='\n')
    {
      line++;
      R= row_alloc(line, -1); insert_row_after(T,R); T= R; Rl= 0; nrows++;
      continue;
    }
    if (Rl==rowW)
    {
      R= row_alloc(line, -1); insert_row_after(T,R); T= R; Rl= 0; nrows++;
      i--;
      continue;
    }
    R->data[Rl]= buf[i];
    Rl++;
  }
  goto again;

end_of_file:
  fclose(f);
  for(;nrows<=pageH;nrows++)
  {
    line++;
    R= row_alloc(line,-1);
    insert_row_after(T,R);
    T= R;
  } 

  if (Crow)
  {
    for(R=Crow;R->prev;R=R->prev) ;
    while(R) { T= R->next; free(R->data); free(R); R= T; }
  }
  rows_head= H;
  Crow= H; Ccol= 0;
  page_layout(H);
  return 0;
}

static void dump_screen()
{
  int i;
  row_t *R;
  for(R=rows_head;R;R=R->next)
  {
   printf("[%3d: %2d] <",R->line, R->drow);
   for(i=0;i<rowW;i++) if (R->data[i]==0) break; else printf("%c",R->data[i]);
   printf(">\n");
  }
}

#define GK(x,y) (((x)<<8)|(y))

void editor_key(int cls, int val)
{
  int i;
  switch(GK(cls,val))
  {
    case GK('C','Q'): return ;
    case GK('C','N'): editor_break_line(); break;
    case GK('C','B'): editor_backspace(); break;
    case GK('A','L'): cursor_move(Crow, Ccol-1); break;
    case GK('A','R'): cursor_move(Crow, Ccol+1); break;
    case GK('A','D'): cursor_move(Crow->next, Ccol); break;
    case GK('A','U'): cursor_move(Crow->prev, Ccol); break;
    case GK('A','H'): cursor_move(Crow, 0); break;
    case GK('P','U'): editor_page_up(); break;
    case GK('P','D'): editor_page_down(); break;
    case GK('A','E'): 
       for(i=rowW-1;i>Ccol;i--) 
         if (Crow->data[i]!=0) { cursor_move(Crow, i); break; }
       break;
    case GK('T',1): dump_screen(); break;
    default:
       if (cls==0) editor_insert_char(val);
       break;
    }
}

void editor_clear()
{
  row_t *R,*H,*T;
  int line;
  int nrows;

  H= T= row_alloc(0, -1);
  line= 0;
  
  for(nrows=1,line=0;nrows<=pageH;nrows++)
  {
    line++;
    R= row_alloc(line,-1);
    insert_row_after(T,R);
    T= R;
  } 

  if (Crow)
  {
    for(R=Crow;R->prev;R=R->prev) ;
    while(R) { T= R->next; free(R->data); free(R); R= T; }
  }
  rows_head= H;
  Crow= H; Ccol= 0;
  page_layout(H);
}
