#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#ifdef GFX_ANDROID
#include <android/native_activity.h>
#include <jni.h>
#endif
#include "gfx.h"
#include "enviro.h"
#include "dbg.h"

char *storage_dir;
int dpi;

#ifdef GFX_ANDROID

static JavaVM *vm= NULL;
static JNIEnv *env;
static int java_ok;

static int java_begin()
{
  int R;
  if (java_ok==1) return 0;
  if (java_ok==2) return 1;
  java_ok= 2;
  
  vm= gfx_get_javavm(); 
  if (!vm) return 1;
  R= (*vm)->AttachCurrentThread(vm, &env, NULL);
  if (R) return 1;
  java_ok= 1;
  return 0;
}

static int java_end()
{
  if (java_ok==2) return 1;
  if (java_ok==0) return 0;
  (*vm)->DetachCurrentThread(vm);
  vm= NULL;
  env= NULL;
  java_ok= 0;
  return 0;
}

static char *get_download_directory()
{
  jmethodID mid;
  jfieldID fid;
  jclass cls;
  jobject obj1,obj2,obj3;
  jboolean iscopy;
  const char *chrs;
  char *k;

#define GPTR(UU) if (!UU) goto done

  k= NULL;
  obj1= obj2= obj3= NULL;

  cls= (*env)->FindClass(env,"android/os/Environment");
  GPTR(cls);
  fid= (*env)->GetStaticFieldID(env, cls,
                   "DIRECTORY_DOWNLOADS", 
                   "Ljava/lang/String;");
  GPTR(fid);
  obj1= (*env)->GetStaticObjectField(env, cls, fid);
  GPTR(obj1);

  mid= (*env)->GetStaticMethodID(env, cls,
                    "getExternalStoragePublicDirectory",
                    "(Ljava/lang/String;)Ljava/io/File;");
  GPTR(mid);

  obj2= (*env)->CallStaticObjectMethod(env, cls, mid, obj1);
  GPTR(obj2);

  cls= (*env)->GetObjectClass(env, obj2);
  GPTR(cls);

  mid= (*env)->GetMethodID(env, cls,
                     "getAbsolutePath", 
                     "()Ljava/lang/String;");
  GPTR(mid);

  obj3= (*env)->CallObjectMethod(env, obj2, mid);
  GPTR(obj3);

  chrs= (*env)->GetStringUTFChars(env, obj3, &iscopy);
  GPTR(chrs);
  k= strdup(chrs);
  (*env)->ReleaseStringUTFChars(env, obj3,chrs);
done:
  if (obj3) (*env)->DeleteLocalRef(env, obj3);
  if (obj2) (*env)->DeleteLocalRef(env, obj2);
  if (obj1) (*env)->DeleteLocalRef(env, obj1);
#undef GPTR
  return k;
}

static char *je_get_str_field(char *fn)
{
  ANativeActivity *activity;
  char *result;
  const char *k;

  jclass activityClass;
  jfieldID fieldID;
  jobject strObject;

  activity= gfx_get_activity();

  activityClass = (*env)->GetObjectClass(env,activity->clazz);
  DPRINT(("after activityClass\n"))
  fieldID = (*env)->GetFieldID(env,activityClass, fn,
                                                "Ljava/lang/String;");
  if (!fieldID) return NULL;
  DPRINT(("after fieldID\n"))
  strObject = (*env)->GetObjectField(env,activity->clazz, fieldID);
  if (!strObject) return NULL;
  k= (*env)->GetStringUTFChars(env,(jstring) strObject, NULL);
  result= strdup(k);
  (*env)->ReleaseStringUTFChars(env, strObject,k);
  (*env)->DeleteLocalRef(env,strObject);
  return result;
}

static void je_call_void_method(char *fn)
{
  ANativeActivity *activity;
  jclass activityClass;
  jmethodID methodID;

  activity= gfx_get_activity();

  activityClass = (*env)->GetObjectClass(env,activity->clazz);
  methodID = (*env)->GetMethodID(env,activityClass, fn, "()V");
  DPRINT(("after methodid '%s' = %p\n", fn, methodID))
  if (!methodID) return ;
  DPRINT(("before the call\n"))
  (*env)->CallVoidMethod(env,activity->clazz, methodID);
  DPRINT(("after the actual call\n"))
}

static int je_call_int_method(char *fn)
{
  ANativeActivity *activity;
  jclass activityClass;
  jmethodID methodID;
  int result;

  activity= gfx_get_activity();

  activityClass = (*env)->GetObjectClass(env,activity->clazz);
  methodID = (*env)->GetMethodID(env,activityClass, fn, "()I");

  if (!methodID) return 0;

  return (*env)->CallIntMethod(env,activity->clazz, methodID);
}

static char *get_storage_directory()
{
  ANativeActivity *activity;
  char *path;
  const char *k;

  jclass activityClass;
  jmethodID getFilesDir;
  jobject fileObject;
  jclass fileClass;
  jmethodID getAbsolutePath;
  jobject pathObject;

  activity= gfx_get_activity();

  if (activity->internalDataPath) return strdup(activity->internalDataPath);

  activityClass = (*env)->GetObjectClass(env,activity->clazz);
  getFilesDir = (*env)->GetMethodID(env,activityClass, "getFilesDir",
                                                "()Ljava/io/File;");
  fileObject = (*env)->CallObjectMethod(env,activity->clazz, getFilesDir);
  fileClass = (*env)->GetObjectClass(env,fileObject);

  getAbsolutePath = (*env)->GetMethodID(env,fileClass, "getAbsolutePath",
                                                "()Ljava/lang/String;");
  pathObject = (*env)->CallObjectMethod(env,fileObject, getAbsolutePath);
  k= (*env)->GetStringUTFChars(env,(jstring)pathObject, NULL);
  path= strdup(k);

  (*env)->ReleaseStringUTFChars(env, pathObject,k);

  (*env)->DeleteLocalRef(env,pathObject);
  (*env)->DeleteLocalRef(env,fileClass);
  (*env)->DeleteLocalRef(env,fileObject);
  (*env)->DeleteLocalRef(env,activityClass);
  
  return path;
}


void init_enviro()
{
  if (!java_begin())
  {
    storage_dir= get_storage_directory();
    gfx_set_sysbar_height(je_call_int_method("getStatusBarHeight"));
    dpi= je_call_int_method("getDisplayDPI");
    if (dpi<40) dpi= 40; // mmmh?
    java_end();
  }
  if (!storage_dir) storage_dir= "/storage/emulated/0/Download"; 
}


#endif

#ifdef GFX_X11

void init_enviro()
{
  storage_dir= strdup("../test");
  dpi= 96;
}
#endif


#ifdef GFX_ANDROID

char *call_sys_str(char *mn, int code, char *fn)
{
  gfx_event_t event;

  if (java_begin()) return NULL;
  
  je_call_void_method(mn);
  DPRINT(("after call_void\n"))
  while(1)
  {
    if (!gfx_event(NULL,&event)) continue;
    if (event.type==GFXEV(code,JAVARESULT))
      return je_get_str_field(fn);
    if (sys_wait_event(&event)) return NULL;
  }
}

void call_sys_void(char *mn)
{
  if (java_begin()) return ;
  je_call_void_method(mn);
}

#endif
