#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include <math.h>
#include <string.h>
#include <dirent.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#ifdef GFX_ANDROID
#include <android/native_activity.h>
#include <jni.h> 
#endif
#include "gfx.h"
#include "editor.h"
#include "gfxeditor.h"
#include "draw.h"
#include "olan.h"

static unsigned int  fgcolor= 0xFFffFF;
static double posx, posy;
static int angle;

static char *fn;

typedef struct { double posx, posy; int angle; } pos_t;

static int npos, spos;
static pos_t *postack;

#define MAGIC1 "@>:-)"
#define MAGIC2 "o-<-<"

void bf_save()
{
  if (npos==spos)
  {
    spos += spos / 2 + 10;
    postack= realloc(postack, sizeof(postack[0])*spos);
  }
  postack[npos].posx= posx;
  postack[npos].posy= posy;
  postack[npos].angle= angle;
  npos++;
}

void bf_restore()
{
  if (npos<1) return ;
  npos--;
  posx= postack[npos].posx;
  posy= postack[npos].posy;
  angle= postack[npos].angle;
}

static int clamp(int value,int L, int G)
{
  if (value<L) return L;
  if (value>G) return G;
  return value;
}

static void color_component(int sh, int val)
{
  val= clamp(val, 0, 100)*255/100;
 
  fgcolor &= ~(0xff<<sh);
  fgcolor |= val<<sh;
  draw_color(fgcolor);
}

void bf_red(int value) { color_component(16, value); }
void bf_green(int value) { color_component(8, value); }
void bf_blue(int value) { color_component(0, value); }
void bf_width(int value) { draw_width(clamp(value,1,20)); }

void bf_right(int value)
{
  angle+= value;
  while (angle<0) angle+= 360;
  angle%= 360; 
}

void bf_left(int value) { bf_right(-value); }

void bf_move(int len)
{
  double alpha;
  alpha= angle/180.0*M_PI;

  posx+= len * sin(alpha);
  posy+= -len*cos(alpha);
}
void bf_line(int len)
{
  double alpha;
  double dx, dy;
  alpha= angle/180.0*M_PI;

  dx= len * sin(alpha);
  dy= -len*cos(alpha);

  draw_line(posx, posy, posx+dx, posy+dy);
  posx= posx+dx;
  posy= posy+dy;
}

void bf_clear()
{
  draw_clear();
}

void bf_tick(int ticks)
{
  gfx_event_t event;
  while(ticks--)
  {
    gfx_timer_start(333, 10);
    while(!gfx_event(NULL, &event) || event.type!=GFXEV(333,TIMEOUT));
  }
}

#ifdef GFX_ANDROID

static JavaVM *vm= NULL;
static JNIEnv *env;
static int java_ok;

static int java_begin()
{
  int R;
  if (java_ok==1) return 0;
  if (java_ok==2) return 1;
  java_ok= 2;

  vm= gfx_get_javavm();
  if (!vm) return 1;
  R= (*vm)->AttachCurrentThread(vm, &env, NULL);
  if (R) return 1;
  java_ok= 1;
  return 0;
}                                        

static int java_end()
{
  if (java_ok==2) return 1;
  if (java_ok==0) return 0;
  (*vm)->DetachCurrentThread(vm);
  vm= NULL;
  env= NULL;
  java_ok= 0;
  return 0;
}

static char *get_download_dir()
{
  jmethodID mid;
  jfieldID fid;
  jclass cls;
  jobject obj1,obj2,obj3;
  jboolean iscopy;
  const char *chrs;
  char *k;

#define GPTR(UU) if (!UU) goto done

  k= NULL;
  obj1= obj2= obj3= NULL;

  cls= (*env)->FindClass(env,"android/os/Environment");
  GPTR(cls);
  fid= (*env)->GetStaticFieldID(env, cls,
                   "DIRECTORY_DOWNLOADS", 
                   "Ljava/lang/String;");
  GPTR(fid);
  obj1= (*env)->GetStaticObjectField(env, cls, fid);
  GPTR(obj1);

  mid= (*env)->GetStaticMethodID(env, cls,
                    "getExternalStoragePublicDirectory",
                    "(Ljava/lang/String;)Ljava/io/File;");
  GPTR(mid);

  obj2= (*env)->CallStaticObjectMethod(env, cls, mid, obj1);
  GPTR(obj2);

  cls= (*env)->GetObjectClass(env, obj2);
  GPTR(cls);

  mid= (*env)->GetMethodID(env, cls,
                     "getAbsolutePath", 
                     "()Ljava/lang/String;");
  GPTR(mid);

  obj3= (*env)->CallObjectMethod(env, obj2, mid);
  GPTR(obj3);

  chrs= (*env)->GetStringUTFChars(env, obj3, &iscopy);
  GPTR(chrs);
  k= strdup(chrs);
  (*env)->ReleaseStringUTFChars(env, obj3,chrs);
done:
  if (obj3) (*env)->DeleteLocalRef(env, obj3);
  if (obj2) (*env)->DeleteLocalRef(env, obj2);
  if (obj1) (*env)->DeleteLocalRef(env, obj1);
#undef GPTR
  return k;
}

static char *get_storage_dir()
{
  ANativeActivity *activity;
  char *path;
  const char *k;

  jclass activityClass;
  jmethodID getFilesDir;
  jobject fileObject;
  jclass fileClass;
  jmethodID getAbsolutePath;
  jobject pathObject;

  if (java_begin()) return strdup("/tmp");

  activity= gfx_get_activity();

  if (activity->internalDataPath) return strdup(activity->internalDataPath);

  activityClass = (*env)->GetObjectClass(env,activity->clazz);
  getFilesDir = (*env)->GetMethodID(env,activityClass, "getFilesDir",
                                                "()Ljava/io/File;");
  fileObject = (*env)->CallObjectMethod(env,activity->clazz, getFilesDir);
  fileClass = (*env)->GetObjectClass(env,fileObject);    

  getAbsolutePath = (*env)->GetMethodID(env,fileClass, "getAbsolutePath",
                                                "()Ljava/lang/String;");
  pathObject = (*env)->CallObjectMethod(env,fileObject, getAbsolutePath);
  k= (*env)->GetStringUTFChars(env,(jstring)pathObject, NULL);
  path= strdup(k);

  (*env)->ReleaseStringUTFChars(env, pathObject,k);

  (*env)->DeleteLocalRef(env,pathObject);
  (*env)->DeleteLocalRef(env,fileClass);
  (*env)->DeleteLocalRef(env,fileObject);
  (*env)->DeleteLocalRef(env,activityClass);

  java_end();

  return path;                        
}
#endif

#ifdef GFX_X11
static char *get_storage_dir()
{
  char *home;
  char buf[2000];
  home= getenv("HOME");
  if (!home) home= "/home/dodo";
  sprintf(buf,"%s/vtackfiles",home);
  return strdup(buf);
}

static char *get_download_dir()
{
  char *home;
  char buf[2000];
  home= getenv("HOME");
  if (!home) home= "/home/dodo";
  sprintf(buf,"%s/Downloads",home);
  return strdup(buf);
}
#endif

static char *storage_dir()
{
  static char *R;
  if (!R) R= get_storage_dir();
  return R;
}

static char *download_dir()
{
  static char *R;
  if (!R) R= get_download_dir();
  return R;
}


static char *fn_to_path(char *fn)
{
  char buf[2000];
  sprintf(buf,"%s/%s",storage_dir(), fn);
  return strdup(buf);
}

#if 1
void gfx_insert(int wno,int pos, char *text,void *user,unsigned int* icon) ;
void gfx_set_text(int wno,char *str);
char* gfx_get_text(int wno);
#endif

static char *import_file_name()
{
  char buf[2000];
  sprintf(buf,"%s/vtack.txt",download_dir());
  return strdup(buf);
}

static void export_file(char *name, FILE *fou)
{
  char buf[2000];
  FILE *fin;
  char *path;
  path= fn_to_path(name);
  fin= fopen(path,"rb");
  if (!fin)
  {
    loprintf("can't open %s: %s\n", path, strerror(errno));
    return ;
  }
  fprintf(fou,"%s %s\n", MAGIC1, name);
  while(fgets(buf,sizeof(buf)-1,fin))
  {
    buf[sizeof(buf)-1]= 0;
    fprintf(fou,"%s",buf);
  }
  fprintf(fou,"%s\n", MAGIC2);
  fclose(fin);
}

static FILE *open_import_file(char *mode)
{
  FILE *ff;
  char *path;
  path= import_file_name();
  ff= fopen(path, mode);
  if (!ff)
    loprintf("can't open %s: %s\n", path, strerror(errno));
  return ff;
}

static void export_files()
{
  FILE *fou;
  DIR *d;
  struct dirent *de;

  fou= open_import_file("wb");
  if (!fou)  return ;

  d= opendir(storage_dir());
  if (!d) goto finish;

  while((de=readdir(d)))
  {
    if (de->d_name[0]=='.') continue;
    export_file(de->d_name,fou);
  }
  closedir(d);
finish:
  fprintf(fou,"END");
  fclose(fou);
}

static int get_line(char *buf, int siz,FILE *fin)
{
  int i;
  if (!fgets(buf,siz-1,fin)) return 'E';
  buf[siz-1]= 0;
  if (!strncmp(buf,MAGIC1,strlen(MAGIC1)))
  {
    char *o;
    for(i=0;buf[i];i++) if (buf[i]=='\n' || buf[i]=='\r') buf[i]= 0;
    for(o=buf+strlen(MAGIC1)+1;*o;o++)
      *(buf++)= *o;
    *buf= 0;
    return 'S';
  }
  if (!strncmp(buf,MAGIC2,strlen(MAGIC2))) return 'F';
  return 'D';
}

static void import_files()
{
  char buf[2000];
  FILE *fin, *fou;
  int q;

  fin= open_import_file("rb");
  if (!fin) return ;
  fou= NULL;

  q= 0;
  while(!q) switch(get_line(buf,sizeof(buf),fin))
  {
  case 'E': q= 1; break;
  case 'S': if (fou) fclose(fou); fou= fopen(fn_to_path(buf),"wb"); break;
  case 'F': if (fou) fclose(fou); fou= NULL; break;
  case 'D': if (fou) fprintf(fou,"%s",buf); break;
  }
  if (fou) fclose(fou);
  fclose(fin);
}

static void load_file()
{
  gfx_SCRload();
  DIR *d;
  int q;
  char *sel_file;
  struct dirent *de;
  d= opendir(storage_dir());
  if (!d) goto list_ok;
  while((de=readdir(d)))
  {
    if (de->d_name[0]=='.') continue;
    gfx_insert(WloadLIST, -1, de->d_name, NULL, NULL);
  }
  closedir(d);
list_ok:
  sel_file= NULL;
  q= 0;
  while(!q)
  {
    gfx_event_t event;
    if (!gfx_event(NULL,&event)) continue;
    switch(event.type)
    {
    case WloadLIST: sel_file= event.string; break;
    case WloadOK: q=1; break;
    case GFXEV_QUIT: 
    case WloadCANCEL: q=2; break;
    }
  }
  if (q==1 && sel_file) sel_file= strdup(sel_file);
  gfx_leave_screen();
  if (q==2 || !sel_file) return ;
  fn= sel_file;
  editor_load(fn_to_path(fn));
  gfx_set_text(Wfn, fn);
}

static void prt_log(const char *data,int size)
{
  gfx_logview_write(Wlog,data,size);
}

static int file_exists(char *ff)
{
  char *p;
  p= fn_to_path(ff);
  if (access(p,F_OK)) return 0;
  return 1;
}

static int save_file()
{
  int q;
  char *ff;
  if (fn)
  {
    editor_save(fn_to_path(fn));
    return 0;
  }
  ff= NULL;
  gfx_SCRsave();
  q= 0;
  while(!q)
  {
    gfx_event_t event;
    if (!gfx_event(NULL,&event)) continue;
    switch(event.type)
    {
    case GFXEV_QUIT:
    case WsaveCANCEL: q=2; break;
    case WsaveOK: 
       ff= gfx_get_text(WsaveFN);
       if (!ff || !ff[0]) continue;
       if (file_exists(ff)) 
       {
           gfx_ask("Hata", "!Tamam", "%s\n adında bir dosya zaten var. "
                                  "Başka bir dosya adı giriniz.",ff);
           continue;
       }
       fn= ff;
       editor_save(fn_to_path(fn));
       q= 1;
       break;
    }
  }
  gfx_leave_screen();
  if (q==1)
  {
    gfx_set_text(Wfn, fn);
    return 0;
  }
  return 1;
}

void bf_circle(int r)
{
  double alpha;
  double dx, dy;
  alpha= angle/180.0*M_PI;

  dx= r*sin(alpha);
  dy= -r*cos(alpha);

  draw_circle(posx+dx, posy+dy, r);
}


static void main_loop()
{
  int r,c;
  gfx_event_t event;
  gfx_SCREEN_0();
  gfxeditor_init(Wedi,&r,&c);
  editor_init(r,c);
  draw_init(Wpix);
  gfx_set_log(Wlog);
//  if (argc>1) editor_load(argv[1]);

  oprint= prt_log;

  draw_width(1);
  draw_color(fgcolor);
  int q;
  q= 0;
  while(!q)
  {
    if (!gfx_event(NULL,&event)) continue;
    if (gfxeditor_event(&event)) continue;
    switch(event.type)
    {
    case GFXEV_QUIT: q= 1; break;
    case Wload: load_file(); break;
    case Wrun: if (save_file()) break;
               orun_file(fn_to_path(fn)); break;
    case Wsave: save_file(); break;
    case Wnew: editor_clear(); fn= NULL; gfx_set_text(Wfn,""); break;
    case Wreset: 
      angle= 0; posx= posy= 0; 
      draw_clear(); gfx_logview_clear(Wlog); 
      break;
    case Wexport: export_files(); loprintf("export finished\n"); break;
    case Wimport: import_files(); loprintf("import finished\n"); break;
    }
  }
}

#ifdef GFX_ANDROID
void android_main(unsigned char *data, size_t len)
{
  main_loop();
}
#else

int main(int argc,char **argv)
{
  gfx_init(NULL, 1140, 895, "Editor Test");
  main_loop();
  return 0;
}
#endif

