#include <sys/types.h> 
#include <setjmp.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdarg.h>
#include <stdint.h> 
#include <ctype.h>

#include "olan.h"

#ifndef BUILTINS
#define BUILTINS 
#endif

/***********************************************************************
 *                                                                     *
 *                          BUILTIN INTERFACE                          *
 *                                                                     *
 ***********************************************************************
  Following macros are used for generating code from builtin function
  declarations.
 ***********************************************************************/

#define __ELEVENTH_ARGUMENT(a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11, ...) a11
#define __COUNT_ARGS(...) \
   __ELEVENTH_ARGUMENT(dummy, ## __VA_ARGS__, 9,8,7,6,5,4,3,2,1,0)

// Now this is basic arithmetic.
#define __INCR_1   2
#define __INCR_2   3
#define __INCR_3   4
#define __INCR_4   5
#define __INCR_5   6
#define __INCR_6   7
#define __INCR_7   8
#define __INCR_8   9
#define __INCR_9   10
#define _INCR(V)  __INCR_##V

#define _COMMA(a,b)  a,b
#define _NOTHING(a,b) a b
//
// COM is a combination macro, either _COMMA or _NOTHING.
// it is used for combining two parts of an iteration
//
#define _ITERC0(F,D,COM)  
#define _ITERC1(F,D,COM,a1) F(a1,D)
#define _ITERC2(F,D,COM,a1,a2) COM(F(a1,D) , F(a2,_INCR(D)))
#define _ITERC3(F,D,COM,a1,...)  \
           COM(F(a1,D) , _ITERC2(F,_INCR(D),COM,__VA_ARGS__))
#define _ITERC4(F,D,COM,a1,...) \
           COM(F(a1,D) , _ITERC3(F,_INCR(D),COM,__VA_ARGS__))
#define _ITERC5(F,D,COM,a1,...) \
           COM(F(a1,D) , _ITERC4(F,_INCR(D),COM,__VA_ARGS__))
#define _ITERC6(F,D,COM,a1,...) \
           COM(F(a1,D) , _ITERC5(F,_INCR(D),COM,__VA_ARGS__))
#define _ITERC7(F,D,COM,a1,...) \
           COM(F(a1,D) , _ITERC6(F,_INCR(D),COM,__VA_ARGS__))
#define _ITERC8(F,D,COM,a1,...) \
           COM(F(a1,D) , _ITERC7(F,_INCR(D),COM,__VA_ARGS__))
#define _ITERC9(F,D,COM,a1,...) \
           COM(F(a1,D) , _ITERC8(F,_INCR(D),COM,__VA_ARGS__))
#define _ITERC10(F,D,COM,a1,...) \
           COM(F(a1,D) , _ITERC9(D),COM,__VA_ARGS__))

#ifndef __CONCAT
#define __CONCAT(a, b) a ## b
#endif
#define __CONCAT2(a, b) __CONCAT(a, b)
#define _FORARGS(F,COM,...) \
  __CONCAT2(_ITERC,__COUNT_ARGS(__VA_ARGS__))(F,1,COM, ##__VA_ARGS__)

#define __TYPEv     void
#define __TYPEi     int
#define __TYPEs     char*
#define _TYPE(r)     __TYPE##r
#define _ARGNAME(D)  a##D
#define _DECL(r,D)   _TYPE(r) _ARGNAME(D)

#define _FUNAME(f) __CONCAT2(bf_,f)

/***********************************************************************
  Builtin function prototypes.
 ***********************************************************************/
#define BF(f,r,...)  _TYPE(r) _FUNAME(f)( _FORARGS(_DECL,_COMMA,##__VA_ARGS__) );
BUILTINS
#undef BF

/***********************************************************************
  Enum definition for builtin functions. 
 ***********************************************************************/
enum { 
 XF_DUMMY,
#define BF(f,r,...) exf_##f,
BUILTINS
#undef BF
};


/***********************************************************************
  Builtin function table.
 ***********************************************************************/
typedef struct { char *name; int number; } bfun_t;

static bfun_t bfuncs[]=
{
#define BF(f,r,...) { #f, exf_##f },
BUILTINS
#undef BF
  { NULL, XF_DUMMY }
};


typedef unsigned char byte_t;

/***********************************************************************
 *                                                                     *
 *                            String Buffer                            *
 *                                                                     *
 ***********************************************************************/
typedef struct strbuf strbuf_t;

struct strbuf {
   struct strbuf *prev,*next;
   char data[4076];
   int len;
};

static void strbuf_puts_l(strbuf_t *sb,const char *str,int len)
{
  if (!sb) return ;
  while(len)
  {
    strbuf_t *t;
    int cp;
    t= sb->prev;
    cp= sizeof(t->data)-t->len;
    if (cp>len) cp= len;
    if (cp==0) 
    {
      t= calloc(1, sizeof(t[0]));
      t->len= 0;
      (t)->prev= (sb)->prev; 
      (t)->next= (sb); 
      (sb)->prev->next= (t); 
      (sb)->prev= (t); 
      continue;
    }
    memcpy(t->data+t->len, str, cp);
    str+= cp;
    len-= cp;
    t->len+= cp;
  }
}

static void strbuf_puts(strbuf_t *sb,char *str)
{
  strbuf_puts_l(sb,str,strlen(str));
}

static void strbuf_vprintf(strbuf_t *sb,const char *fmt,va_list ap)
{
  const char *raw;
  char *sarg; int iarg; char carg;
  int k,sgn;
  char buf[40];

again:
  for(raw=fmt;raw[0];raw++) if (raw[0]=='%') break;
  if (raw!=fmt) strbuf_puts_l(sb, fmt, raw-fmt);
  if (raw[0]==0) return ;
  switch(raw[1])
  {
  case 0: strbuf_puts(sb, "%"); return ;
  case 'd':
    iarg= va_arg(ap,int);
    if (iarg==0) { strbuf_puts_l(sb,"0",1); break; }
    k=sizeof(buf)-1;
    buf[k--]= 0;
    if (iarg<0) { iarg= -iarg; sgn= 1; } else sgn= 0;
    while(iarg) { buf[k--]= iarg%10+'0'; iarg/=10; }
    if (sgn) buf[k--]= '-';
    strbuf_puts(sb,buf+k+1);
    break; 
  case 's':
    sarg= va_arg(ap,char*);
    strbuf_puts(sb,sarg);
    break; 
  case 'c':
    carg= va_arg(ap,int);
    strbuf_puts_l(sb, &carg, 1);
    break;
  default:
    strbuf_puts_l(sb, raw+1, 1);
    break;
  }
  fmt= raw+2;
  goto again;
}

static void strbuf_printf(strbuf_t *sb,const char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  strbuf_vprintf(sb,fmt,ap);
  va_end(ap);
}


static strbuf_t* strbuf_init()
{
  strbuf_t *t;
  t= calloc(1, sizeof(t[0]));
  t->len= 0;
  t->prev= t->next= t;
  return t;
}

static int sb_collect(strbuf_t *sb, char *dst) {
  strbuf_t *k;
  int L;
  k= sb;
  L= 0;
  do {
    if (dst) memcpy(dst+L, k->data, k->len);
    L+= k->len;
    k= k->next;
  } while(k!=sb);
  return L;
}

static void sb_free(strbuf_t *sb) 
{
  while(sb->prev!=sb)
  { 
    strbuf_t *t= sb->prev;
    sb->prev= t->prev;
    free(t);
  }
  free(sb);
}

static char *strbuf_collect(strbuf_t *sb)
{
  int l;
  char *k;
  l= sb_collect(sb, 0);
  k= malloc( l + 1);
  sb_collect(sb, k);
  k[l]= 0;
  sb_free(sb);
  return k;
}

/***********************************************************************
 *                                                                     *
 *                        Basic I/O and Errors                         *
 *                                                                     *
 ***********************************************************************/

static char *c_fn; static int c_line; char *c_func;
static jmp_buf ejbuf;

void (*oerr)(char *fn,int line);
void (*oprint)(const char *buf, int size);

static void oprtv(const char *fmt, va_list ap);
static void oprt(const char *fmt,...);

static int verbose=0;

typedef struct pos { char *fn; char *func;
                     int line; struct pos *prev; } pos_t;
static pos_t *callstack;

static void print_stack(int all,int limit);

static void out_pos(char *fn,int line, char *func)
{
  if (func && func[0]) oprt("%s(%s:%d)", func, fn,line);
                  else oprt("%s:%d",fn,line);
}

static int p_pos(pos_t *K)
{
  if (!K->prev) return 0;
  if (p_pos(K->prev)) oprt("->");
  out_pos(K->fn,K->line,K->func);
  return 1;
}

static void dump_callstack()
{
  oprt("callstack: ");
  if (p_pos(callstack)) oprt("->");
  out_pos(c_fn,c_line,c_func);
  oprt("\n");
}

static void callstack_push(char *fn,int line,char *func)
{
  pos_t *p;
  p= malloc(sizeof(*p));
  p->fn= c_fn; p->line= c_line; p->func= c_func;
  p->prev= callstack;
  callstack= p;
  c_fn= fn;
  c_line= line;
  c_func= func;
}

static void callstack_destroy()
{
  pos_t *P;
  while(callstack)
  {
    P= callstack->prev;
    free(callstack);
    callstack= P;
  }
}

static void callstack_pop()
{
  pos_t *p;
  if (!callstack) return ;
  p= callstack;
  callstack= p->prev;
  c_fn= p->fn;
  c_line= p->line;
  c_func= p->func;
  free(p);
}

static void errbegin(char *fn,int ln)
{
  if (oerr)  (*oerr)(fn,ln); 
}

static void errend()
{
  oprt("\n");
  dump_callstack();
  print_stack(1,10);
  longjmp(ejbuf,1);
}

static void p_errorl(int line, const char *fmt, ...)
{
  va_list ap;
  errbegin(c_fn, line);
  va_start(ap,fmt);
  oprtv(fmt,ap);
  va_end(ap);
  errend();
}

static void p_error(const char *fmt,...)
{
  va_list ap;
  errbegin(c_fn, c_line);
  va_start(ap,fmt);
  oprtv(fmt,ap);
  va_end(ap);
  errend();
}

static void ierror(const char *fmt,...)
{
  va_list ap;
  errbegin("internal",1);
  va_start(ap,fmt);
  oprtv(fmt,ap);
  va_end(ap);
  errend();
}

static void *return_kf(void *a, void *b)
{
  free(b); 
  return a;
}

static char *comstr(char *a,char *b)
{
  char *R;
  R= malloc(strlen(a)+strlen(b)+1);
  sprintf(R,"%s%s",a,b);
  return R;
}

static char *comstr_kf(char *a, char *b) { return return_kf(comstr(a,b), b); }
static char *comstr_ff(char *a, char *b) {return return_kf(comstr_kf(a,b),a);}
static char *comstr_fk(char *a, char *b) { return return_kf(comstr(a,b),a); }

static void init_pos()
{
  c_fn= "ENTRY"; c_line= 1; c_func= "";
}



/***********************************************************************
 *                                                                     *
 *                             String Map                              *
 *                                                                     *
 ***********************************************************************/
typedef struct {
  void *root;
} strmap_t;

static strmap_t *strmap_new() __attribute__((unused));
static void *strmap_find(strmap_t *map, char *key);
static void *strmap_insert(strmap_t *map, char *key,void *data);
static void *strmap_remove(strmap_t *map, char *key);
static void *strmap_destroy(strmap_t *map);

typedef struct
{
  void* child[2];
  unsigned internal:1, byte:23, mask:8;
} strmap_node_t;


static inline int strmap_dir
    (uint8_t *key,uint32_t len_key,strmap_node_t *node)
{
  uint8_t c;

  if (node->byte<len_key) c= key[node->byte];
                     else c= 0;
  return (1+(node->mask|c))>>8;
}

static void* strmap_find(strmap_t *map,char *skey)
{
  uint8_t *key; unsigned int len_key;
  strmap_node_t *node;

  node= map->root;
  if (!node) return 0;

  key= (void*) skey;
  len_key= strlen(skey);

  while(node->internal)
     node= node->child[strmap_dir(key,len_key,node)];

  if (strcmp(skey, node->child[0])) return 0;
                               else return node->child[1];
}


static inline uint32_t strmap_make_mask
    (uint32_t A, uint32_t B)
{
  uint8_t diff;
  diff= A ^ B;
  diff|= diff>>1;
  diff|= diff>>2;
  diff|= diff>>4;
  return (diff&~(diff>>1))^255;
}

static inline strmap_node_t* strmap_leaf_node
    (uint8_t *key, void *data)
{
  strmap_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 0;
  N->child[0]= key;
  N->child[1]= data;
  return N;
}

static inline strmap_node_t* strmap_internal_node
    (uint32_t byte, uint32_t mask)
{
  strmap_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 1;
  N->byte= byte;
  N->mask= mask;
  return N;
}

static void* strmap_insert(strmap_t* map, char* skey, void *data)
{
  uint8_t *key; uint32_t len_key;
  strmap_node_t *node_elt;

  key= (void*) skey;

  if (!map->root)
  {
     map->root= strmap_leaf_node(key,data);
     return 0;
  }

  len_key= strlen(skey);

  node_elt= map->root;
  while(node_elt->internal)
    node_elt= node_elt->child[strmap_dir(key,len_key,node_elt)];

  uint32_t crit_byte;
  uint8_t  mask;
  uint8_t  *elt;
  int dir_elt;
  strmap_node_t *node_key;

  elt= node_elt->child[0];

  for(crit_byte=0;
      crit_byte<=len_key && elt[crit_byte]==key[crit_byte];
      crit_byte++)
          ;

  if (crit_byte>len_key) return node_elt->child[1];

  mask= strmap_make_mask(elt[crit_byte],key[crit_byte]);

  dir_elt= (1+(mask|elt[crit_byte]))>>8;

  node_key= strmap_internal_node(crit_byte,mask);
  node_key->child[1-dir_elt]= strmap_leaf_node(key,data);

  strmap_node_t *parent;
  strmap_node_t *node;

  parent= 0;
  node= map->root;

  while(node->internal)
  {
     if (node->byte > crit_byte) break;
     if (node->byte==crit_byte && node->mask > mask) break;
     parent= node;
     node= node->child[strmap_dir(key,len_key,node)];
  }

  node_key->child[dir_elt]= node;
  if (parent) { if (parent->child[0]==node) parent->child[0]= node_key;
                                    else    parent->child[1]= node_key; }
  else map->root= node_key;
  return 0;
}

static void *strmap_remove(strmap_t* map, char *skey)
{
  uint8_t *key; uint32_t len_key;
  void *retv;

  if (!map->root) return 0;

  key= (void*) skey;
  len_key= strlen(skey);
   
  strmap_node_t *node;
  strmap_node_t *sibling;
  strmap_node_t *parent;
  strmap_node_t *gparent;

  node= map->root;
  gparent= 0;
  sibling= 0;
  parent= 0;

  while(node->internal)
  {
    int dir;
    dir= strmap_dir(key, len_key, node);
    gparent= parent;
    parent= node;
    sibling= node->child[1-dir];
    node= node->child[dir];
  }

  if (strcmp(node->child[0],skey)) return 0;
  retv= node->child[1];
  free(node);

  if (!gparent)
  {
     map->root= sibling;
     if (parent) free(parent);
     return retv;
  }

  if (gparent->child[0]==parent)
    gparent->child[0]= sibling;
  else
    gparent->child[1]= sibling;

  free(parent);

  return retv;
}

static void *strmap_destroy(strmap_t *map)
{
  strmap_node_t *node;
  strmap_node_t *parent;
  strmap_node_t *gparent;
  strmap_node_t *sibling;
  void *retv;

  if (!map->root) return 0;

  parent= 0;
  gparent= 0;
  sibling= 0;
  node= map->root;
  while(node->internal)
  {
    gparent= parent;
    parent= node;
    sibling= node->child[1];
    node= node->child[0];
  }
  retv= node->child[1];
  free(node);
  if (!gparent) 
  {
    map->root= sibling;
    if (parent) free(parent);
    return retv;
  }
  if (gparent->child[0]==parent)
    gparent->child[0]= sibling;
  else
    gparent->child[1]= sibling;

  free(parent);

  return retv;
}

static strmap_t  *strmap_new()
{
  strmap_t *s;
  s= malloc(sizeof(*s));
  s->root= 0;
  return s;
}




/***********************************************************************
 *                                                                     *
 *                        Stack Cell and Values                        *
 *                                                                     *
 ***********************************************************************/
typedef struct cell cell_t;
typedef struct node node_t; 
typedef struct list list_t;

enum { CELL_MARK_BOS=1, CELL_MARK_QUOT,
       CELL_MARK_FILE,  /*** When a file is included, this is pushed on
                        the stack. At the end of the file, the stack is
                        popped until this marker. ***/
       CELL_INT, CELL_FLO,
       CELL_STR, CELL_LIST };
struct cell
{
  int type;
  union { char *vstr; list_t *vlis; double vflo; int vint; } V;
};

struct node { cell_t value; struct node *prev, *next; }; 
struct list { node_t head; }; 

#define CINT(x) (x).V.vint
#define CFLO(x) (x).V.vflo
#define CLIS(x) (x).V.vlis
#define CSTR(x) (x).V.vstr

static list_t *list_new()
{
  list_t *L;
  L= malloc(sizeof(*L));
  L->head.next= L->head.prev= &L->head;
  return L;
}

static void list_append_node(list_t *L, node_t *N)
{
  N->prev= L->head.prev;
  N->next= &L->head;
  N->prev->next= N;
  L->head.prev= N;
}


static void list_append(list_t *L, cell_t *K)
{
  node_t *N;
  N= malloc(sizeof(*N));
  N->value= K[0];
  list_append_node(L,N);
}

static void cell_destroy(cell_t *C);

static int list_empty(list_t *L)
{
  return L->head.prev==&L->head;
}

static int list_remove_first(list_t *L, cell_t *C)
{
  node_t *N;
  if (list_empty(L)) return 0;
  N= L->head.next;
  N->prev->next= N->next;
  N->next->prev= N->prev;
  *C= N->value;
  free(N);
  return 1;
}

static void free_list(list_t *L)
{
  cell_t V;
  while(list_remove_first(L,&V)) cell_destroy(&V);
  free(L);
}

static void cell_destroy(cell_t *C)
{
  switch(C->type)
  {
  case CELL_STR: free(CSTR(*C)); CSTR(*C)= NULL; break;
  case CELL_LIST: free_list(CLIS(*C)); CLIS(*C)= NULL; break;
  }
}

static void cell_copy(cell_t *R, cell_t *A)
{
  list_t *OL, *NL; // original, new list
  node_t *ON;      // original node

  *R= *A;
  if (R->type==CELL_STR) { CSTR(*R)= strdup(CSTR(*A)); return ; }
  if (R->type!=CELL_LIST) return ;

  OL= CLIS(*A);
  CLIS(*R)= NL= list_new();
  for(ON=OL->head.next;ON!=&OL->head;ON=ON->next)
  {
    cell_t V;
    cell_copy(&V, &ON->value);
    list_append(NL, &V);
  }
}

static void cell_empty(cell_t *R)
{
  R->type= CELL_STR;
  CSTR(*R)= strdup("");
}


#if 0
static void list_concat(list_t *H, list_t *T)
{
  node_t *TF, *TL; // T first, last
  node_t *HL;      // H last

  if (list_empty(T)) goto done;
  TF= T->head.next;
  TL= T->head.prev;
  if (TF==TL) { list_append_node(H,TF); goto done; }
  if (list_empty(H))
  {
     H->head.next= TF; TF->prev= &H->head;
     H->head.prev= TL; TL->next= &H->head;
     goto done;
  }
  HL= H->head.prev;

  HL->next= TF; TF->prev= HL;
  TL->next= &H->head; H->head.prev= TL;
done:
  free(T);
}
#endif


static void cell_tostr_d(cell_t *C, strbuf_t *B)
{
  node_t *K;
  list_t *L;
  if (C->type==CELL_STR) { strbuf_puts(B,CSTR(*C)); goto done; } 
  if (C->type!=CELL_LIST) return ;
  L= CLIS(*C);
  for(K=L->head.next;K!=&L->head;K=K->next)
  {
    if (K!=L->head.next) strbuf_puts(B," "); 
    cell_tostr_d(&K->value,B);
  }
done:
  cell_destroy(C);
}

static int cell_is_true(cell_t *C)
{
  switch(C->type)
  {
  case CELL_STR: return toupper(CSTR(*C)[0])=='T'; 
  case CELL_LIST: return !list_empty(CLIS(*C));
  case CELL_INT: return CINT(*C)!=0;
  case CELL_FLO: return CFLO(*C)!=0.0;
  }
  return 0;
}

static int cell_to_int(cell_t *C,int destroy)
{
  int R;
  switch(C->type)
  {
  case CELL_STR: R= atoi(CSTR(*C)); break;
  case CELL_INT: R= CINT(*C); break;
  case CELL_FLO: R= CFLO(*C); break;
  default: R= 0;
  }
  if (destroy) cell_destroy(C);
  return R;
}

static char* cell_to_str(cell_t *C,int destroy)
{
  char* R;
  char buf[40];
  switch(C->type)
  {
  case CELL_STR: R= strdup(CSTR(*C)); break;
  case CELL_INT: sprintf(buf,"%d",CINT(*C)); R= strdup(buf); break;
  case CELL_FLO: sprintf(buf,"%f",CFLO(*C)); R= strdup(buf); break;
  default: R= strdup("[]");
  }
  if (destroy) cell_destroy(C);
  return R;
}

/***********************************************************************
 *                                                                     *
 *                              The Stack                              *
 *                                                                     *
 ***********************************************************************/

static cell_t *stack;
static int stack_size;
static int stack_tos, stack_bos;


static void init_stack()
{
  stack_size= 100;
  stack=malloc(sizeof(*stack)*stack_size);
  stack_tos= 0;
  stack_bos= -1;
}

static void push(cell_t C)
{
  if (stack_tos==stack_size)
  {
    stack_size+= stack_size/2;
    stack= realloc(stack, sizeof(*stack)*stack_size);
  }
  stack[stack_tos]= C;
  stack_tos++;
}

static int pop(cell_t *C,int force)
{
  cell_t *K;
  if (stack_tos==0) goto underflow;
  K= stack+stack_tos-1;
  if (K->type==CELL_MARK_FILE ||
      K->type==CELL_MARK_BOS || K->type==CELL_MARK_QUOT)
     goto underflow;
  *C= *K;
  stack_tos--;
  return 0;
underflow:
  if (force) p_error("stack underflow");
  return 1;
}

static void push_word(byte_t *s,byte_t *e)
{
  cell_t C;
  char *V;
  C.type= CELL_STR;
  V= malloc(e-s+1);
  memcpy(V,s,e-s);
  V[e-s]= 0;
  CSTR(C)= V;
  push(C);
}

static void push_marker(int m)
{
  cell_t C;
  C.type= m;
  push(C);
}

static void print_cell(cell_t *C,int eol)
{
  node_t *K;
  switch(C->type)
  {
  case CELL_STR: oprt("s(%s)", CSTR(*C)); break;
  case CELL_MARK_FILE: oprt("FILE"); break;
  case CELL_MARK_QUOT: oprt("QUOT"); break;
  case CELL_MARK_BOS: oprt("BOS"); break;
  case CELL_LIST: oprt("("); 
   for(K=CLIS(*C)->head.next;K!=&CLIS(*C)->head;K=K->next)
     print_cell(&K->value,0); 
   oprt(")");
   break;
  default: oprt("?");
  }
  if (eol) oprt("\n");
}

static void print_stack(int all,int limit)
{
  int e,i;
  oprt("stack(tos=%d, bos=%d):\n",stack_tos,stack_bos);
  if (all) e= -1; else e= stack_bos;
  if (limit>0) { if (stack_tos-limit>e) e=stack_tos-limit; }
  for(i=stack_tos-1;i>e;i--) print_cell(stack+i,1);
  oprt("end\n");
}

static void pop_to(int marker)
{
  int i;
  for(i=stack_tos-1;i>stack_bos;i--)
     if (stack[i].type==marker) break;
     else cell_destroy(stack+i);
  if (i>stack_bos) cell_destroy(stack+i);
  stack_tos= i;
}

#if 0
static void push_list(list_t *L)
{
  cell_t C;
  C.type= CELL_LIST;
  CLIS(C)= L;
  push(C);
}
#endif

static void pop_str_force(cell_t *C,char* op)
{
  pop(C,1); 
  if (C->type!=CELL_STR) p_error("argument to %s must be a string",op);
}
void push_str(char *K) { cell_t C; C.type=CELL_STR; CSTR(C)= K; push(C); }
void push_strc(char *K) { push_str(strdup(K)); }
void push_int(int K) { cell_t C; C.type= CELL_INT; CINT(C)=K; push(C); }

/***********************************************************************
 *                                                                     *
 *                              Variables                              *
 *                                                                     *
 ***********************************************************************
 When a variable is set, the value is lifted of the stack and stored as
 is. When we get the value of a variable, a deep copy of the value is
 made. The functions in this file do this copy. There is no need to do
 it outside.
 ***********************************************************************/

typedef struct scope { strmap_t vars; struct scope *prev; } scope_t;
typedef struct { char *name; cell_t *value; } var_t;

scope_t *glob_scope, *file_scope, *func_scope;

static scope_t *scope_new()
{
  scope_t *s;
  s= calloc(1,sizeof(*s));
  return s;
}
static scope_t *scope_destroy(scope_t *s)
{
  scope_t *P;
  var_t *V;
  P= s->prev;
  while((V=strmap_destroy(&s->vars)))
  {
    cell_destroy(V->value);
    free(V->value);
    free(V->name);
    free(V);
  }
  return P;
}

static var_t *var_find_scope(char *str, scope_t *sc)
{
  return strmap_find(&sc->vars, str);
}

static var_t *var_find(char *str)
{
  var_t *V;
  if (func_scope && (V=var_find_scope(str,func_scope))) return V;
  return var_find_scope(str,glob_scope);
}

static int var_get(char *str, cell_t *RC)
{
  var_t *V;
  if (!strcmp(str,"empty"))
  {
    RC->type= CELL_LIST;
    CLIS(*RC)= list_new();
    return 0;
  }
  if ((V=var_find(str)))
  {
    cell_copy(RC, V->value);
    return 0;
  }
  return 1;
}


static int var_set(char *str, cell_t *C)
{
  var_t *V;

  V= var_find(str);
  if (V)
  {
    cell_destroy(V->value);
  }
  else
  {
    V= malloc(sizeof(*V));
    V->name= strdup(str);
    V->value= malloc(sizeof(V->value[0]));
    strmap_insert(&func_scope->vars, V->name, V);
  }

  // careful: make a shallow copy. the value comes from the stack
  // and it will be forgotten about pretty soon.
  V->value[0]= C[0];
  return 0;
}

static void var_destroy(var_t *V)
{
  free(V->name);
  cell_destroy(V->value);
  free(V->value);
  free(V);
}

static void var_setforce_glob(char *name, char *value)
{
  var_t *V;
  if (func_scope && (V=strmap_remove(&func_scope->vars,name)))
   var_destroy(V);
  if ((V=strmap_remove(&glob_scope->vars,name)))
   var_destroy(V);
  V= malloc(sizeof(*V));
  V->value= malloc(sizeof(*V));
  V->value->type= CELL_STR;
  CSTR(V->value[0])= value;
  V->name= name;
  strmap_insert(&glob_scope->vars,name,V);
}
/***********************************************************************
 *                                                                     *
 *                          Function Storage                           *
 *                                                                     *
 ***********************************************************************/
#define FUNC_BUILTIN  1
#define FUNC_USER     2

typedef struct
{
  char *name;
  int type;
  void (*builtin)();
  byte_t *body;
  byte_t *end;
  char *d_file; int d_line;
} func_t;

static strmap_t funcs;

static void func_builtin(char *name, void (*f)())
{
  func_t *F;
  if (strmap_find(&funcs,name))
   ierror("function %s is already defined",name);
  F= calloc(1,sizeof(*F));
  F->type= FUNC_BUILTIN;
  F->name= strdup(name);
  F->builtin= f;
  strmap_insert(&funcs, name, F);
}


/***********************************************************************
 *                                                                     *
 *                        PARSING AND EXECUTION                        *
 *                                                                     *
 ***********************************************************************/

/***********************************************************************
 The following type is used for the escape character table.
 ***********************************************************************/
typedef struct { char *name, *value; } namval_str_t;



static void p_getvar()
{
  cell_t C,V;
  pop_str_force(&C,"$");
  if (var_get(CSTR(C),&V)) cell_empty(&V);
  cell_destroy(&C);
  push(V);
}

static void p_setvar()
{
  cell_t C,V;
  pop_str_force(&C,"=");
  if (pop(&V,0)) cell_empty(&V); 
  var_set(CSTR(C), &V);
  cell_destroy(&C);
}

static void p_escape()
{
  int i;
  cell_t C; char *str;
  static namval_str_t esctab[]=
  {
    { "pl", "+"  }, { "pc", "%"  }, { "ls", "["  }, { "rs", "]"  },
    { "lc", "{"  }, { "rc", "}"  }, { "eq", "="  }, { "dl", "$"  },
    { "at", "@"  }, { "ha", "#"  }, { "qm", "?"  }, { "qu", "\"" },
    { "bs", "\\" }, { "nl", "\n" }, { "sp", " "  }, { "tb", "\t" },
    { NULL, NULL },
  };
  pop_str_force(&C,"%");
  str= "";
  for(i=0;esctab[i].name;i++) 
    if (!strcmp(esctab[i].name,CSTR(C)))
    {
       str= esctab[i].value;
       break;
    }
  cell_destroy(&C);
  push_strc(str);
}

#if 0
static void p_plus()
{
  cell_t AC,LC;
  list_t *L;
  pop(&AC,1);
  if (pop(&LC,0)) 
  {
    L= list_new();
    list_append(L, &AC);
    push_list(L);
    return ;
  }
  if (LC.type==CELL_LIST) L= CLIS(LC);
                     else list_append(L=list_new(),&LC);
  if (AC.type==CELL_LIST) list_concat(L, CLIS(AC));
                     else list_append(L, &AC);
  push_list(L);
}
#endif

static void p_quot()
{
  int s,i;
  cell_t C;
  strbuf_t *B;
  for(s=stack_tos-1;s>stack_bos;s--)
    if (stack[s].type==CELL_MARK_QUOT)
      break;
  if (s==stack_bos) p_error("unmatched ]");
  B= strbuf_init();
  for(i=s+1;i<stack_tos;i++)
  {
    if (i!=s+1) strbuf_puts(B," ");
    cell_tostr_d(stack+i,B);
  }
  stack_tos= s;
  C.type= CELL_STR;
  CSTR(C)= strbuf_collect(B);
  push(C);
}

static byte_t *p_curly(byte_t *p, byte_t *e)
{
  byte_t *t;
  int sline, qline;
  int k;

  sline= c_line;
  p++;
  t= p; 
  k= 1;

again:
  if (p>=e) p_errorl(sline,"unmatched {"); 
  switch(*p)
  {
  case '{': k++; break;
  case '}': k--; if (k==0) { push_word(t,p); return p+1; } break;
  case '\n': c_line++; break;
  case '?': for(;p<e;p++) if (p[0]=='\n') break; goto again;
  case '\"': 
    qline= c_line; 
    for(p++;p<e;p++)
      if (p[0]=='\"' && p[-1]!='\\') break; else if (p[0]=='\n') c_line++;
    if (p>=e) p_errorl(qline, "unmatched \""); 
    break;
  }
  p++; 
  goto again;
}

static byte_t *p_string(byte_t *p,byte_t *e)
{
  byte_t *t;
  int sline;
  t= p;
  sline= c_line;
  p++;

again:
  if (p>=e) p_errorl(sline,"unmatched \""); 
  switch(*p)
  {
  case '\n': c_line++; break;
  case '\"': if (p[-1]!='\\') { push_word(t,p+1); return p+1; }
  }
  p++; 
  goto again;
}


static void process_chars(byte_t *p,byte_t *e);

static int pop_int_force(char *op)
{
  cell_t C;
  pop(&C,1); 
  return cell_to_int(&C,1);
}

static void p_binarith(int op)
{
  int a,b,c;
  char buf[20];
  buf[0]= op; buf[1]= 0;
  b= pop_int_force(buf); 
  a= pop_int_force(buf);
  switch(op)
  {
  case '+': c= a+b; break;
  case '-': c= a-b; break;
  case '*': c= a*b; break;
  case '/': if (b) c= a/b; else c= 0; break;
  case '%': if (b) c= a%b; else c= 0; break;
  }
  sprintf(buf,"%d",c);
  push_strc(buf);  
}

static void p_add() { p_binarith('+'); }
static void p_sub() { p_binarith('-'); }
static void p_div() { p_binarith('/'); }
static void p_mul() { p_binarith('*'); }
static void p_mod() { p_binarith('%'); }

static int gptr;

static void pop_stack(char *op, int num)
{
  if (stack_tos<num) 
  {
    p_error("stack underflow in %s",op);
    return ;
  }
  stack_tos-= num;
  gptr= stack_tos;
}
static int get_int() __attribute__((unused));

static int get_int() 
{
  return cell_to_int( stack+gptr++, 1);
}

static char *get_str()
{
  return cell_to_str(stack+gptr++, 1);
}

//
// this is for builtins defined by the user of the library
//
static int call_builtin(char *name)
{
#define DEFU(x)  __attribute__((unused)) int x;
  DEFU(Ai1) DEFU(Ai2) DEFU(Ai3) DEFU(Ai4) DEFU(Ai5) 
  DEFU(Ai6) DEFU(Ai7) DEFU(Ai8) DEFU(Ai9) DEFU(Ai10)  DEFU(ri)
#undef DEFU
#define DEFU(x) char *x __attribute__((unused));
  DEFU(As1) DEFU(As2) DEFU(As3) DEFU(As4) DEFU(As5) 
  DEFU(As6) DEFU(As7) DEFU(As8) DEFU(As9) DEFU(As10) DEFU(rs)
#undef DEFU
  int i;
  for(i=0;bfuncs[i].name;i++)
   if (!strcmp(bfuncs[i].name,name))
   {
     switch(bfuncs[i].number)
     {
 // get the return value.
#define __GETVi  ri= 
#define __GETVs  rs= 
#define __GETVv  
#define _GETV(t) __GETV##t

// get argument from the stack.
#define __GETAi(D)  Ai##D = get_int();
#define __GETAs(D)  As##D = get_str();  
#define _GETA(t,D) __GETA##t(D)

// put argument into function call
#define __PUTA(t,D)  A##t##D
#define _PUTA(t,D)  __PUTA(t,D)

// push the result back on the stack
#define __PUSHVi  push_int(ri);
#define __PUSHVs  push_str(rs);
#define __PUSHVv  
#define _PUSHV(t) __PUSHV##t

#define BF(f,r,...)  \
  case exf_##f: \
    pop_stack(#f,__COUNT_ARGS(__VA_ARGS__)); \
    _FORARGS(_GETA,_NOTHING,##__VA_ARGS__) \
    _GETV(r) _FUNAME(f)( _FORARGS(_PUTA,_COMMA,##__VA_ARGS__) ); \
    _PUSHV(r) \
  break;
BUILTINS
#undef BF

     }
     return 1;
   }
  return 0;
}

static void p_call()
{
  cell_t C;
  func_t *F;
  scope_t *s_func_scope;

  pop_str_force(&C,"#");
  if (call_builtin(CSTR(C))) return ;
  F= strmap_find(&funcs,CSTR(C));
  if (!F) p_error("unknown function %s",CSTR(C));
  cell_destroy(&C);
  if (F->type==FUNC_BUILTIN) { (*F->builtin)(); return ; }

  callstack_push(F->d_file, F->d_line, F->name);
  s_func_scope= func_scope; func_scope= scope_new();
  process_chars(F->body,F->end);
  scope_destroy(func_scope);
  callstack_pop();
  func_scope= s_func_scope;
}

static void process_chars(byte_t *p,byte_t *e)
{
  static byte_t ccl[256];
  byte_t *t;

  if (!ccl['a'])
  {
    const char *special= "+-*/%[]{}=$@#?\"";
    int i;
    for(i=0;i<256;i++) ccl[i]= 'w';
    for(i=0;i<=' ';i++) ccl[i]='s'; ccl['\n']='n'; ccl[127]= 's';
    for(i=0;special[i];i++) ccl[(unsigned int)special[i]]= special[i];
  }

again:
  if (p>=e) goto done;
  switch(ccl[*p])
  {
  case 's': break;
  case 'n': c_line++; break;
 
  case '-':
  case '+': 
     if (p+1==e || ccl[*(p+1)]!='w')
     {
        if (ccl[*p]=='-') p_sub(); else p_add(); 
        p++;
        goto again;
     } 
  case 'w': for(t=p,p++;p<e&&ccl[*p]=='w';p++) ; push_word(t,p); goto again;
  case '?': for(t=p;p<e&&*p!='\n';p++) ;  goto again;
  case '$': p_getvar(); break;
  case '=': p_setvar(); break;
  case '@': p_escape(); break;
 
  case '/': p_div(); break;
  case '%': p_mod(); break;
  case '*': p_mul(); break;

  case '[': push_marker(CELL_MARK_QUOT); break;
  case ']': p_quot(); break;
  case '{': p= p_curly(p,e); goto again;
  case '}': p_error("unmatched }"); break;
  case '"': p= p_string(p,e); break;
  case '#': p_call(); break;
  default: p_error("internal error: unclassified character %c",*p); 
  }
  p++; 
  goto again;
done:
  return ;
}

static void process_chunk(cell_t *C)
{
  // do something similar to p_call, but don't put these at the
  // call stack. these get executed as part of an if# statement etc.
  // in fact, we probably don't need the s_fn, but I'll include it
  // anyways, just to make sure.
  char *s_fn; int s_line;
  char *body;
  s_fn= c_fn; s_line= c_line; 
  body= CSTR(*C);
  process_chars((byte_t*)body,(byte_t*)body+strlen(body));
  c_fn= s_fn; c_line= s_line;
}

static int process_mem(char *fn, byte_t *p, byte_t *e)
{
  scope_t *s_file_scope, *s_func_scope;
  int R;
  R= 0;

  callstack_push(fn,1,"");
  s_file_scope= file_scope; s_func_scope= func_scope;

  c_fn= fn; c_line= 1; file_scope= func_scope= scope_new();
  push_marker(CELL_MARK_FILE);

  if (setjmp(ejbuf)) { R= 1; goto done; }
  process_chars(p,e);

done:
  while(func_scope!=file_scope) func_scope= scope_destroy(func_scope);
  scope_destroy(file_scope);
  file_scope= s_file_scope; func_scope= s_func_scope;
  callstack_pop();
  pop_to(CELL_MARK_FILE);
  return R;
}


/***********************************************************************
 *                                                                     *
 *                          BUILTIN FUNCTIONS                          *
 *                                                                     *
 ***********************************************************************/

static void b_print()
{
  cell_t C;
  pop(&C,1);
  if (verbose<0) 
  {
    cell_destroy(&C);
    return ;
  }
  if (C.type==CELL_STR)
  {
    oprt("%s\n",CSTR(C));
    cell_destroy(&C);
  }
  else
  {
    strbuf_t *B;
    char *S;
    B= strbuf_init();
    cell_tostr_d(&C,B);
    S= strbuf_collect(B);
    oprt("%s\n",S);
    free(S);
  }
}

void b_dump() { dump_callstack(); print_stack(1,10); }

static void b_define()
{
  cell_t N;
  cell_t B;
  char *name, *body;
  func_t *F;
  pop_str_force(&N,"define"); name= CSTR(N); 
  pop_str_force(&B,"define"); body= CSTR(B);
  F= strmap_find(&funcs, name);
  if (F)
  {
    if (F->type==FUNC_BUILTIN) 
      p_error("attempt to redefine builtin function %s", name);
#if 0
    p_warn("function %s defined at %s:%d is redefined at %s:%d\n",
            name, F->d_file, F->d_line, c_fn, c_line);
#endif
    free(F->body);
    free(name);
  }
  else
  {
    F= calloc(1,sizeof(*F));
    F->name= name;
    F->type= FUNC_USER;
    strmap_insert(&funcs, name, F);
  }
  F->body= (byte_t*) body; F->end= (byte_t*)body+strlen(body);
  F->d_file= c_fn; F->d_line= c_line;
}

static void b_if()
{
  cell_t body,cond;
  pop_str_force(&body, "if");
  pop_str_force(&cond, "if");
  if (cell_is_true(&cond)) process_chunk(&body);
  cell_destroy(&body);
  push(cond);
}

static void b_else()
{
  cell_t body,cond;
  pop_str_force(&body, "else");
  pop_str_force(&cond, "else");
  if (!cell_is_true(&cond)) process_chunk(&body);
  cell_destroy(&body);
  push(cond);
}

static void b_drop() { cell_t C; pop(&C,1); cell_destroy(&C); }

static list_t *make_wrapper_list(cell_t *C,int *Rde)
{
  list_t *L;
  if (C->type==CELL_LIST) { *Rde=0; return CLIS(*C); }
  L= list_new();
  list_append(L,C);
  *Rde= 1;
  return L;
}

// this is a special function which deallocates a list
// with a single node. the value in the node is not
// destroyed, just the jig holding it.
static void delete_wrapper_list(int de,list_t *L)
{
  if (!de) return ;
  free(L->head.next);
  free(L);
}



static void b_rep_str()
{
  char *p,*nn,*z, *rep, *srch, *vv;
  int repl, srchl;
  cell_t *V;
  int R;
  cell_t S, SRCH, REP;
  list_t *L; node_t *N; int de;
  pop_str_force(&REP,"rep.str"); rep= CSTR(REP); repl= strlen(rep);
  pop_str_force(&SRCH,"rep.str"); srch= CSTR(SRCH); srchl= strlen(srch);
  pop(&S,1);
  R= 0;
  L= make_wrapper_list(&S, &de);
  for(N=L->head.next;N!=&L->head;N=N->next)
  {
    V= &N->value;
    if (V->type!=CELL_STR) p_error("bad argument to rep.str");
    vv= CSTR(*V);
    z= strstr(vv,srch);
    if (!z) continue;
    nn= malloc(strlen(vv)-strlen(srch)+strlen(rep)+1);
    p= nn;
    memcpy(p, vv, z-vv); p+= z-vv;
    memcpy(p, rep, repl); p+= repl; 
    strcpy(p, z+srchl);
    free(CSTR(*V));
    CSTR(*V)= nn;
    R++;
  }
  if (de) S= L->head.next->value;
  delete_wrapper_list(de,L);
  push(S);
  push_strc(R?"True":"False");
}

static void b_cat()
{
  list_t *L; node_t *N; 
  cell_t A,B;

  pop(&B,1); pop(&A,1);

  switch(A.type*256+B.type)
  {
  case CELL_STR*256+CELL_STR:
    CSTR(A)= comstr_ff(CSTR(A),CSTR(B));
    break;
  case CELL_STR*256+CELL_LIST:
    L= CLIS(B);
    for(N=L->head.next;N!=&L->head;N=N->next)
      if (N->value.type!=CELL_STR) p_error("bad arg to cat");
      else CSTR(N->value)= comstr_kf(CSTR(A),CSTR(N->value));
    cell_destroy(&A);
    A= B;
    break;
  case CELL_LIST*256+CELL_STR:
    L= CLIS(A);
    for(N=L->head.next;N!=&L->head;N=N->next)
      if (N->value.type!=CELL_STR) p_error("bad arg to cat");
      else CSTR(N->value)= comstr_fk(CSTR(N->value),CSTR(B));
    cell_destroy(&B);
    break;
  default: 
    p_error("bad args to cat");
  }
  push(A);  
}

static void b_export()
{
  var_t *V,*VV;
  cell_t N;
  char *name;
  pop_str_force(&N,"export");
  name= CSTR(N);
  V= NULL;
  if (func_scope && (V=strmap_find(&func_scope->vars,name)))
  {
     strmap_remove(&func_scope->vars,name);
     VV= strmap_remove(&glob_scope->vars,name);
     if (VV) var_destroy(VV);
     strmap_insert(&glob_scope->vars,V->name,V);
     free(name);
  }
  else
  {
     VV= strmap_find(&glob_scope->vars,name);
     if (VV) { free(name); return ; }
     V= malloc(sizeof(*V));
     V->name= name;
     V->value= malloc(sizeof(*V->value));
     cell_empty(V->value);
     strmap_insert(&glob_scope->vars,name,V);
  }
}


static void b_getenv()
{
  cell_t C;
  char *s;
  pop_str_force(&C,"getenv");
  s= getenv(CSTR(C));
  cell_destroy(&C);
  if (s) push_str(s); else push_strc("");
}

static void equality_comparison(int sense)
{
  cell_t A,B;
  int r;
  pop_str_force(&B,"eq");
  pop_str_force(&A,"eq");
  r= strcmp(CSTR(A),CSTR(B)) ? !sense : sense ;
  cell_destroy(&A);
  cell_destroy(&B);
  push_strc(r?"True":"False");
}

static void b_eq() { return equality_comparison(1); }
static void b_neq() { return equality_comparison(0); }


static void b_foreach()
{
  list_t *L; node_t *N; int de;
  cell_t LC, BODY;
  pop_str_force(&BODY,"for.each");
  pop(&LC,1);
  L= make_wrapper_list(&LC, &de);
  for(N=L->head.next;N!=&L->head;N=N->next)
  {
    push(N->value);
    process_chunk(&BODY);
  }
  delete_wrapper_list(de, L);
}

static void b_times()
{
  cell_t IDX, BODY;
  int i,N;
  pop_str_force(&IDX,"times");
  pop_str_force(&BODY,"times");

  N= atoi(CSTR(IDX));
  for(i=0;i<N;i++)
    process_chunk(&BODY);
}

static void glob_scope_init()
{
  if (!glob_scope) glob_scope= scope_new();
}

static void init_builtins()
{
  func_builtin("print", b_print);
  func_builtin("dump", b_dump);
  func_builtin("define", b_define);
  func_builtin("if", b_if);
  func_builtin("else", b_else);
  func_builtin("drop", b_drop);
  func_builtin("rep.str", b_rep_str);
  func_builtin("cat", b_cat);
  func_builtin("export", b_export);
  func_builtin("getenv", b_getenv);
  func_builtin("eq", b_eq);
  func_builtin("neq", b_neq);
  func_builtin("for.each", b_foreach);
  func_builtin("times", b_times);
}

int orun(char *fn, unsigned char *start, unsigned char *end)
{
  int R;
  func_t *F;
  glob_scope_init();
  init_stack();
  init_pos();
  init_builtins();

  R= process_mem(fn, start, end);
  
  while((F=strmap_destroy(&funcs)))
  {
    if (F->type==FUNC_USER) free(F->body);
    free(F->name);
    free(F);
  }
  stack_size= 0;
  stack_tos= 0;
  stack_bos= 0;
  free(stack);
  stack= NULL;

  callstack_destroy();

  return R;
}

int orun_file(char *fn)
{
  strbuf_t *B;
  FILE *f;
  char *data;
  char buf[2000];
  int R;
  f= fopen(fn,"rb");
  if (!f)
  {
    oprt("'%s' dosyası açılamadı\n", fn);
    return 1;
  }
  B= strbuf_init();
  while((R=fread(buf,1,sizeof(buf),f)))
   strbuf_puts_l(B, buf, R);
  fclose(f);
  data= strbuf_collect(B);
  R= orun(fn,(unsigned char*)data, (unsigned char*)data+strlen(data));
  free(data);
  return R;
}

#if 0
/***********************************************************************
 *                                                                     *
 *                            PROGRAM ENTRY                            *
 *                                                                     *
 ***********************************************************************/
static void cmdline_define_var(char *name,char* value)
{
  var_setforce_glob(name,value);
}

static char *input_files[200]; static int n_input_files;

static void parse_options(int argc,char **argv)
{
  int i, j, sh;
  char *name,*value;
again:
  if (argc==1) return ;
  sh= 1;
  if (argv[1][0]=='-')
  {
    switch(argv[1][1])
    {
     case 'V': printf("jj version 3.0 2019-05-22\n"); exit(1);
     case 'q': verbose= -1; break;
     case 'v': verbose= 1; break;
     case 'f': if (argc<3) goto fail;
               input_files[n_input_files++]= argv[2];
               sh= 2; break;
     default: goto fail;
    }
  }
  else if (strchr(argv[1],'='))
  {
     name= strdup(argv[1]);
     value= strchr(name,'='); *value= 0; value++;
     cmdline_define_var(name,value);
  } 
  else
  {
    cmdline_define_var(strdup(argv[1]),strdup("True"));
  }

  for(i=1+sh,j=1;i<argc;i++,j++) argv[j]= argv[i];
  argc-= sh;
  goto again;
fail:
  fprintf(stderr,
"%s: bad command line. possible arguments, all optional:\n"
"  -V                      print version and exit\n"
"  -f input_file_name      can be given multiple times\n"
"  -q                      quiet mode, doesn't even print messages\n"
"  -v                      verbose, prints commands before executing them\n"
"  var=value               predefine variable\n"
"  var                     predefine variable=True\n"
   , argv[0]);
  exit(1);
}

void echo(char *s)
{
  printf("[%s]\n",s);
}

int main(int argc,char **argv)
{
  int i;
  glob_scope= scope_new();
  init_stack();
  init_pos();
  init_builtins();
  parse_options(argc,argv);
  if (!n_input_files) input_files[n_input_files++]= "opi";
  for(i=0;i<n_input_files;i++)
    process_file(NULL,input_files[i]);
  return 0;
}
#endif

static void o_out(const char *buf, int size)
{
  if (oprint) { (*oprint)(buf,size); return ; }
  fwrite(buf, 1,size, stdout);
}

static void o_outs(char *str)
{
  o_out(str, strlen(str));
}

static void oprtv(const char *fmt, va_list ap)
{
  const char *raw;
  char *sarg; int iarg; char carg;
  int k,sgn;
  char buf[40];

again:
  for(raw=fmt;raw[0];raw++) if (raw[0]=='%') break;
  if (raw!=fmt) o_out(fmt, raw-fmt);
  if (raw[0]==0) return ;
  switch(raw[1])
  {
  case 0: o_outs("%"); return ;
  case 'd':
    iarg= va_arg(ap,int);
    if (iarg==0) { o_outs("0"); break; }
    k=sizeof(buf)-1;
    buf[k--]= 0;
    if (iarg<0) { iarg= -iarg; sgn= 1; } else sgn= 0;
    while(iarg) { buf[k--]= iarg%10+'0'; iarg/=10; }
    if (sgn) buf[k--]= '-';
    o_outs(buf+k+1);
    break; 
  case 's':
    sarg= va_arg(ap,char*);
    o_outs(sarg);
    break; 
  case 'c':
    carg= va_arg(ap,int);
    o_out(&carg, 1);
    break;
  default:
    o_out(raw+1, 1);
    break;
  }
  fmt= raw+2;
  goto again;
}

static void oprt(const char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  oprtv(fmt,ap);
  va_end(ap);
}

void *__olan_suppress_unused[]= { strbuf_printf , var_setforce_glob, };
